-- Miscellaneous routines

global constant DOS32 = 1, -- ex.exe
		WIN32 = 2  -- exw.exe

constant M_PLATFORM = 53

global function platform()
-- return the platform (operating system) we are running on
    return machine_func(M_PLATFORM, 0)
end function

constant M_INSTANCE = 55

global function instance()
-- WIN32: returns hInstance - handle to this instance of the program
-- DOS32: returns 0
    return machine_func(M_INSTANCE, 0)
end function

global function reverse(sequence s)
-- reverse the top-level elements of a sequence.
-- Thanks to Hawke' for helping to make this run faster.
    integer lower, n
    sequence t
    
    n = length(s)
    t = repeat(0, n)
    lower = 1
    for upper = n to floor(n/2)+1 by -1 do
	t[upper] = s[lower]
	t[lower] = s[upper]
	lower += 1
    end for
    return t
end function

global function sprint(object x)
-- Return the string representation of any Euphoria data object. 
-- This is the same as the output from print(1, x) or '?', but it's
-- returned as a string sequence rather than printed.
    sequence s
				 
    if atom(x) then
	return sprintf("%.10g", x)
    else
	s = "{"
	for i = 1 to length(x) do
	    s &= sprint(x[i])  
	    if i < length(x) then
		s &= ','
	    end if
	end for
	s &= "}"
	return s
    end if
end function

-- trig formulas provided by Larry Gregg

global constant PI = 3.141592653589793238

constant PI_HALF =  PI / 2.0  -- this is pi/2

type trig_range(object x)
--  values passed to arccos and arcsin must be [-1,+1]
    if atom(x) then
	return x >= -1 and x <= 1
    else
	for i = 1 to length(x) do
	    if not trig_range(x[i]) then
		return 0
	    end if
	end for
	return 1
    end if
end type

global function arccos(trig_range x)
--  returns angle in radians
    return PI_HALF - 2 * arctan(x / (1.0 + sqrt(1.0 - x * x)))
end function

global function arcsin(trig_range x)
--  returns angle in radians
    return 2 * arctan(x / (1.0 + sqrt(1.0 - x * x)))
end function
