-----------------------------------------------------------------------------
-- ZIG Object Execution Engine ----------------------------------------------
-- Rewrite 2: Compiling version ---------------------------------------------
-- ZIG v1.0 -----------------------------------------------------------------
-- (C) 1999, JacobH / RSC. --------------------------------------------------
-- Website: http://surf.to/zig ----------------------------------------------
-- E-mails: hammond@humboldt1.com - for general contacting ------------------
----------- zig16@hotmail.com - ZIG-specific --------------------------------
-----------------------------------------------------------------------------


-- Euphoria includes --------------------------------------------------------
include graphics.e
include image.e
include file.e
include get.e
include sort.e
include misc.e

-- Custom includes ----------------------------------------------------------
include modwave.e
include palette.e
include keyread.e


without trace
with    type_check
without warning

allow_break(1)


-- Global constants: --------------------------------------------------------

global constant DEBUGGER = 0, DEBUG_FILE = 0,    -- debug constants
                
                UNSUP_CMD = 999,                 -- unsupported command const
                
                -- messages
                MAX_MESSAGES = 100, MSG_MSG = 1, MSG_COLOR = 2,
                MSG_TIME = 3, MSG_X = 4, MSG_Y = 5, CENTERED = 0,

                -- directional constants
                DIR_N = 220, DIR_S = 221, DIR_E = 222, DIR_W = 223
                

-- Global variables: --------------------------------------------------------

global atom SoundFreq, SoundOn, SoundChannels, SoundStereo,
            CColorF, CColorB, CColor, exits, TPS,
            NextAvailableMessage,
            
            MessageX, MessageY, MessageColor, MessageTime

global sequence CustTiles,   -- never used, incl for compat with zsidebar.e
                DefaultPalette, DefaultCharSet, CharSet,
                
                CompileData, CWName,
                
                PreCommands,
                
                VarVals, VarNames, FlagNames, FlagVals, Run, VisibleLayers,
                
                Messages, ScrollBuffers,
                
                SideBar


-- ZIG shared-code includes
include zchar.e
include zlibrs.e
include zobjects.e
include zboards.e
include zgfx.e
include zchared.e
include zgui.e
include zworlds.e
include zutil.e
include zfades.e


-- Variable setup -----------------------------------------------------------

-- Define boards (OneBoard stuck onto/into Boards for new board)
OneBoard = repeat(0, BOARD_ELEMS)
Boards = {OneBoard}

-- World and board properties, using W_PROPS / B_PROPS
WorldProps = repeat({}, length(W_PROPS))
BoardProps = repeat(repeat({}, length(B_PROPS)), 1)

-- Object definitions, and object position table
           -- boardn,x,y,layer,char,col,dmode,pos,state,locked,cycle,name,hasp,props x 4
OneObject = {{  0,   0,0,  0,   0,  00,   0,   0,   0,     0,    0,   {},  0,  0,0,0,0}}
Objects = {}
ObjectPosTable = repeat(repeat(repeat(0, 3), 80), 25)

-- Setup flags and variables
FlagNames = {}      --
FlagVals = {}       -- parallels
VarNames = {}       -- 
VarVals = {}        --

-- Messages setup
Messages = repeat({{}, 0, 2, 1, 0, 0}, MAX_MESSAGES)

-- Make data for compiler
CompileData = {
    {"#_focus", 1},             -- sets/clears this object as a main 'focus' obj
    {"#become", UNSUP_CMD},     -- unsupported
    {"#bind", 2},               -- copy @obj's program to this object
    {"#change", 3},             -- change all item x to item y on board
    {"#char", 4},               -- change the object's character to n
    {"#clear", 5},              -- clear flag flagname
    {"#clearflag", 5},          -- same as #clear
    {"#clone", 6},              -- make a clone of @obj at (x,y,[layer],[board])
    {"#color", 7},              -- change object's color to n
    {"#cycle", 8},              -- set the object's cycle speed to n
    {"#die", 9},                -- destroy the object
    {"#end", 10},               -- stop running for now
    {"#endgame", 11},           -- ends the current running game
    {"#give", 12},              -- give item itemname amount n
    {"#go", 13},                -- go in direction dir
    {"#idle", 14},              -- be idle for (cycle * tickspeed) duration
    {"#if", 15},                -- conditional - #if (boolean) then (command, [command..]) [else (command, [command..])]
    {"#loadfont", 16},          -- loads font file and sets it as current set
    {"#loadmod", 17},           -- load a music module
    {"#playmod", 18},           -- start a mod playing
    {"#pausemod", 19},          -- pause the currently playing mod
    {"#move", 20},              -- move object to x,y[,layer][,board]
    {"#playwav", 21},           -- plays wave file if a channel is available
    {"#lock", 22},              -- lock the object
    {"#redraw", 23},            -- redraw the current board, and all objects
    {"#restart", 24},           -- restart the object
    {"#restore", 25},           -- restore (un-zap) zapped label
    {"#unzap", 25},             -- same as #restore
    {"#scroll", 26},            -- scroll entire board w/ absolute dirs (x,y)
    {"#send", 27},              -- jump to label in program
    {"#set", 28},               -- set a flag
    {"#setflag", 28},           -- same as #set
    {"#set_msx", 29},           -- set message position x
    {"#set_msy", 30},           -- set message position y
    {"#set_mscolor", 31},       -- set message color (0 - flashing)
    {"#set_mstime", 32},        -- set message duration (cycles)
    {"#take", 33},              -- take item itemname amount n
    {"#transport", 34},         -- move to board n (any _focus objects moved)
    {"#try", 13},               -- equiv to #go
    {"#unlock", 35},            -- unlock object
    {"#visiblelayers", 36},     -- set layers visibility to [a,b,c] boolean
    {"#vislayers", 36},         -- same as #visiblelayers, but
    {"#walk", 13},              -- equiv to #go
    {"#zap", 37},               -- change a label to 'zapped'
    {"#variable", 38},          -- define a new variable varname, type vtype
    {"#transportnof", 39},      -- same as #transport, but don't move _focus objs
    {"#scrollarea", 40},        -- #scroll, (x,y,x1,y1,x2,y2)
    {"#changearea", 41},        -- #change, (stype,dtype,x1,y1,x2,y2)
    {"#addstat", 42},           -- add a status item to the sidebar (use #give/#take)
    {"#delstat", 43},           -- remove a status item
    {"#volume", 44},            -- set the master volume
    {"#volup", 45},             -- slide the master volume up 1
    {"#voldown", 46},           -- slide the master volume down 1
    
    {"/n", 101},                -- move north [n] space
    {"/s", 102},                -- move south [n] space
    {"/e", 103},                -- move east [n] space
    {"/w", 104},                -- move west [n] space
    {"?n", 101},                -- \/ same as /n /s /e /w
    {"?s", 102},                --
    {"?e", 103},                --
    {"?w", 104},                --
    {"/i", 105},                -- same as #idle
    {"\\n", 106},               -- /n no auto /i (\\ is esc char, really '\n')
    {"\\s", 107},               -- /s no auto /i
    {"\\e", 108},               -- /e no auto /i
    {"\\w", 109},               -- /w no auto /i
    
    {"%me.x", 201},             -- object's x position
    {"%me.y", 202},             -- object's y position
    {"%me.layer", 203},         -- object's layer
    {"%curboard", 204},         -- current board
    {"%prop1", 205},            -- obj property 1
    {"%prop2", 206},            -- "          " 2
    {"%prop3", 207},            -- "          " 3
    {"%prop4", 208},            -- "          " 4
    {"alligned", 209},          -- is obj alligned with @obj
    {"getobjname", 210},        -- gets object name at (x,y)
    {"keyb", 211},              -- key condition of key (str)
    {"blocked", 212},           -- is object blocked at dir
    {"seek", 213},              -- dop: direction towards object [@obj] def: @Player
    {"rnd", 214},               -- dop: random direction
    {"rndns", 215},             -- dop: random dir, n/s
    {"rndew", 216},             -- dop: random dir, e/w
    {"cw", 217},                -- dop: turn dir clockwise
    {"ccw", 218},               -- dop:          counterclockwise
    {"opp", 219},               -- dop: opposite of direction dir
    {"n", 220},                 -- dop: north (sync with DIR_% constants)
    {"s", 221},                 -- dop: south
    {"e", 222},                 -- dop: east
    {"w", 223},                 -- dop: west
    
    {"+", 301},                 -- add two numbers
    {"-", 302},                 -- sub
    {"*", 303},                 -- mul
    {"/", 304},                 -- div
    {"=", 305},                 -- comparison, or assignment (with %%varname)
    
    {"else", 401},              -- used in #if
    
    {"integer", 501},           -- data type for #variable
    {"string", 502}             -- not yet supported
    
    }
    
function GetConfig()
    atom f
    sequence config
    
    if length(dir("zconfig.inf")) != 0 then
        f = open("zconfig.inf", "r")
        config = get(f)
        if length(config) != 0 then
            config = config[2]
        else SoundOn = FALSE return 1
        end if
        close(f)
        if length(config) > 1 then
            SoundOn = TRUE
            SoundFreq = str2num(config[2])
            SoundChannels = 16
            if config[3] = 1 then
                SoundStereo = STEREO
            else SoundStereo = MONO
            end if
        else
            SoundOn = FALSE
        end if
    else
        SoundOn = FALSE
        return 1
    end if
    
    return 0
end function

procedure Init()
    atom t
    sequence dx
    
    if graphics_mode(3) then
        puts(1, "\ngraphics_mode(3) returned error!")
        EndProg(ERR_INIT_ERROR)
    end if
    
    color(15, 1)
    puts(1, " Init(): ZIG Game Engine Initialization ")    
    t = time() + .5
    while time() < t do
    end while

    color(7, 0)
    
    puts(1, "\n\nZIG Game Engine, version 2.0\nCopyright 1997-99, JacobH/RSC.\nInitializing...\n")
    
    puts(1, "\ngraphics_mode(): Display initialized, mode 3.\n")
    
    t = GetConfig()
    if t then
        puts(1, "GetConfig(): Returned error!\n")
    else puts(1, "GetConfig(): Configuration read OK.\n")
    end if

    dx = {}
    for z = 0 to 15 do
        dx = append(dx, get_palette(z))
    end for
    puts(1, "BIOS palette read OK.\n")
    
    puts(1, "Switching videomode...\n")
    t = time() + .5
    while time() < t do
    end while
    
    if DEBUGGER then
        t = text_rows(28)
        puts(1, "In-game debugger is on, so egachars() not called.\n(640x480x16 screen, 9x14 charbox, 80x28).\n")
    else
        egachars()
        puts(1, "egachars(): EGA characterset mode initialized.\n(640x350x16 screen, 8x14 charbox, 80x25).\n")
        puts(1, "In-game debugger is off.\n")
    end if
    
    
    allow_all_paledit()
    puts(1, "allow_all_paledit(): Palette patch installed.\n")
    cursor(NO_CURSOR)
    puts(1, "cursor(): Cursor reset (NO_CURSOR).\n")
    toggle_blinking(0)
    puts(1, "toggle_blinking(): 16x16 colormode enabled.\n")
    ZIGSwitchPal(dx)
    puts(1, "ZIGSwitchPal(): Default BIOS palette initialized.\n")
    DefaultPalette = dx
    ZIGLoadFont("default.zch")
    if length(CharSet) != 0 then
        DefaultCharSet = CharSet
        puts(1, "ZIGLoadFont(): OK, default characterset loaded and initialized.\n")
    else
        puts(1, "ZIGLoadFont(): returned bad characterset.\nThe default charset file is missing or corrupted!!\nUsing default BIOS font in place of custom set.\n")
        t = wait_key()
        t = load_rom_font()
        DefaultCharSet = CharSet
    end if
    sound(0)
    puts(1, "sound(): Reset.\n")
    
    BoardProps[1][BP_CHARSET] = DefaultCharSet
    BoardProps[1][BP_PAL] = DefaultPalette
    
    if SoundOn = TRUE then
        if init_modwave(SoundStereo, SoundFreq, SoundChannels) then
            puts(1, "init_modwave(): Error initializing sound driver!\nTry using lower settings, mono mode, or less channels.\nOr, try turning the sound off.\n")
            SoundOn = FALSE
        else
            printf(1, "init_modwave(): Sound engine initialized:\n\tFrequency:\t%d\n\tChannels:\t%d\n\tStereo:\t\t%d\n", {SoundFreq, SoundChannels, SoundStereo})
        end if
    else
        puts(1, "init_modwave(): Not initializing sound driver (sound is turned off).\n")
    end if
    
    tick_rate(100)
    puts(1, "tick_rate(): High-resolution timering enabled.\n")
    

    CurBoard = 1
    
    Scroll_SizeX = 16
    Scroll_SizeY = 50
    Scroll_PosX = 5
    Scroll_PosY = 10
    Scroll_TextC = 14
    Scroll_BackC = 1
    Scroll_HLTitleC = 15
    Scroll_BorderC = 9
    
    NextAvailableMessage = 1    
    MessageX = 25
    MessageY = 0
    MessageTime = 2
    MessageColor = 0
        
    TPS = .1
    
    t = time() + .5
    while time() < t do
    end while
    
end procedure

procedure GameError(sequence Text, sequence Cap)
    atom k
    sequence ss
    
    ss = SaveScreen()
    
    Box(8, 10, 14, 70, 4, Cap)
    if not atom(Text[1]) then
        for z = 1 to length(Text) do
            position(9 + z, 12)
            puts(1, Text[z])
        end for
    else
        position(10, 12)
        puts(1, Text)
    end if
    
    position(12, 12)
    text_color(14)
    puts(1, "Press any key...")
    k = wait_key()
    clear_keys()
    
    ResScreen(ss)
end procedure

procedure SelectWorld(atom m)
    atom rs
    sequence fin
    
    fin = ChooseFile("*.zig")
    
    if fin[1] = 1 and length(fin) = 1 then
        if m then
            EndProg(0)
        else
            return
        end if
    else
        rs = LoadWorld(fin)
        if rs = 0 then
            CWName = fin
        else
            GameError("An error occured while loading the world!", "Error!")
            abort(1)                    
        end if
    end if
end procedure

procedure InitWorld()
    atom tb, zz
    sequence SS, Words, CObjects
    object dum, dum2
    
    SS = SaveScreen()
    
    if length(Objects) = 0 then
        GameError("This world contains no objects!", "Error while preparing world")
        EndProg(1)
    end if
    
    Box(8, 5, 16, 55, 1, "World Initialization")
    
    position(10, 7)
    printf(1, "Compiling %d objects, please wait...", length(Objects))
    
    position(12, 7)
    puts(1, repeat(176, 25))
    
    CObjects = {}
    text_color(14)
    
    for z = 1 to length(Objects) do
        position(12, 7)
        puts(1, repeat(219, (z * 50) / (length(Objects) * 2)))
        position(13, 7)
        printf(1, "Object %d", z)
        CObjects = append(CObjects, {Objects[z][1]})
        for a = 2 to length(Objects[z]) do
            position(14, 7)
            printf(1, "%5d/%5d: ", {a, length(Objects[z])})
            if length(Objects[z][a]) > 34 then
                puts(1, Objects[z][a][1..34])
            else puts(1, Objects[z][a] & repeat(32, 35 - length(Objects[z][a])))
            end if
            
            CObjects[z] = append(CObjects[z], {})
            
            if length(Objects[z][a]) = 0 then
                CObjects[z][a] = {900} & {" "}
            else
                if Objects[z][a][1] = '\'' then
                    CObjects[z][a] = {}
                elsif Objects[z][a][1] = '@'
                or Objects[z][a][1] = ':'
                or Objects[z][a][1] = '|'
                or Objects[z][a][1] = '%'
                or Objects[z][a][1] = '#'
                or Objects[z][a][1] = '/'
                or Objects[z][a][1] = '\\' then
                
                    dum = Objects[z][a]
                    dum2 = 0
                    while dum2 = 0 do
                        if find(',', dum) != 0 then
                            dum = dum[1..find(',', dum) - 1] & dum[find(',', dum) + 1..length(dum)]
                        else dum2 = 1
                        end if
                    end while
                    
                    Words = {}
                    dum2 = {} 
                    for c = 1 to length(dum) do
                        if dum[c] = ' '
                        or c = length(dum) then
                            if c = length(dum) then dum2 = dum2 & dum[c] end if
                            Words = append(Words, dum2)
                            dum2 = {}
                        else
                            dum2 = dum2 & dum[c]
                        end if
                    end for
                    
                    if Objects[z][a][1] = '#'
                    or Objects[z][a][1] = '/'
                    or Objects[z][a][1] = '%'
                    or Objects[z][a][1] = '\\' then
                        for c = 1 to length(Words) do
                            dum = {}
                            for d = 1 to length(CompileData) do
                                if compare(Words[c], CompileData[d][1]) = 0 then
                                    dum = CompileData[d][2]
                                    exit
                                else
                                    dum = Words[c]
                                end if
                            end for
                            CObjects[z][a] = append(CObjects[z][a], dum)
                        end for
                    else
                        for c = 1 to length(Words) do
                            CObjects[z][a] = append(CObjects[z][a], Words[c])
                        end for
                    end if
                else
                    CObjects[z][a] = {900} & {Objects[z][a]}
                end if            
            end if
        end for
    end for
    
    Words = Objects
    Objects = CObjects
    CObjects = Words

    if length(WorldProps[WP_TBOARD]) != 0 then
        CurBoard = WorldProps[WP_TBOARD][1]
    else CurBoard = 1
    end if
    
    ResScreen(SS)
end procedure

function BuildRunTable(atom b)
    sequence r
    
    r = {}
    
    for z = 1 to length(Objects) do
        if length(Objects[z]) != 0 then
            if Objects[z][1][OBJ_BOARD_NUM] = b
            or Objects[z][1][OBJ_BOARD_NUM] = -1 then  -- global
                r = append(r, z)
            end if
        end if
    end for
    
    return(r)
end function

procedure dispcmd(sequence a)
    for z = 1 to length(a) do
        if sequence(a[z]) then
            if compare(a[z], {0}) = 0 then
                puts(1, "0")
            else puts(1, a[z] & ' ')
            end if
        else puts(1, a[z] & ' ')
        end if
    end for
    puts(1, '\n')
end procedure

procedure DrawGameBar(atom typeof, sequence conts)
    sequence ar

    SB_Clear()
    SB_CText(2, SB_HLITE1, " - - ZIG - - ")
    SB_CText(24, 7, "v1.0")
    
    if typeof = 1 then  -- titlebar
        SB_BtnText(5, 3, " W  Current World:")
        text_color(15)
        bk_color(0)
        position(6, 66)
        puts(1, " " & CWName & " ")
        SB_BtnText(9, 7, " P  Play World")
        SB_BtnText(10, 3, " L  Load Game")
        SB_BtnText(12, 7, " A  About ZIG")
        SB_BtnText(13, 3, " H  Help")
        SB_BtnText(14, 7, " Q  Quit")
        SB_Text(21, 7, "ZIG Game Engine")
        SB_Text(22, 7, "version 1.0.")
        SB_Text(23, 7, "(C) 1999, Jacob H.")
        SB_Text(24, 7, "ZIG is freeware.")
    else
        bk_color(1)
        for z = 1 to length(conts) do
            position(5 + z, 62)
            text_color(15)
            puts(1, conts[z][3])
            text_color(14)
            position(5 + z, 64)
            puts(1, conts[z][2])
            if sequence(conts[z][4]) then
                ar = conts[z][4][2..length(conts[z][4])]
                for zz = 1 to length(VarNames) do
                    if compare(ar, VarNames[zz]) then
                        printf(1, ": %d", VarVals[zz])
                        exit
                    end if
                end for
            end if
        end for
    end if
end procedure

procedure InitWorld_2()
    Run = BuildRunTable(CurBoard)
    BuildObjectTable(CurBoard)
    PreCommands = repeat({}, length(Run))
    ScrollBuffers = repeat({}, length(Run))

    if length(BoardProps[CurBoard][BP_CHARSET]) != 0 then
        CharSet = BoardProps[CurBoard][BP_CHARSET]
        add_all_fonts(0, CharSet)
    else
        CharSet = DefaultCharSet
        add_all_fonts(0, CharSet)
    end if
    
    if length(BoardProps[CurBoard][BP_PAL]) != 0 then
        ZIGSwitchPal(BoardProps[CurBoard][BP_PAL])
    else ZIGSwitchPal(DefaultPalette)
    end if
    
    VisibleLayers = {1, 1, 1}
    
    SideBar = {}

    DrawBoard(CurBoard, VisibleLayers)

    DrawObjectsInGame(Run, repeat({0,0,0,1}, length(Run)), CurBoard)
    
end procedure

function FindObject(sequence Name, sequence Runz)
    atom z
    
    Name = trim(Name)
    if length(Runz) != 0 then
        for ax = 1 to length(Runz) do                
            z = Runz[ax]
            if compare(Objects[z][1][OBJ_NAME], Name) = 0 then
                return(z)
            end if
        end for
    end if
    
    return(FALSE)
end function

procedure NewMessage(sequence Mess)
    Messages[NextAvailableMessage][MSG_MSG] = Mess[1]
    Messages[NextAvailableMessage][MSG_COLOR] = Mess[2]
    Messages[NextAvailableMessage][MSG_TIME] = time() + (Mess[3] * TPS)
    Messages[NextAvailableMessage][MSG_X] = Mess[4]
    Messages[NextAvailableMessage][MSG_Y] = Mess[5]
    NextAvailableMessage = NextAvailableMessage + 1
end procedure
             
procedure UpdateMessages()
    atom Why

    for z = 1 to NextAvailableMessage - 1 do
        if length(Messages[z][MSG_MSG]) != 0 then
            if Messages[z][MSG_TIME] < time() then
                DrawBoard1L(CurBoard, Messages[z][MSG_X], VisibleLayers)
                if z = NextAvailableMessage - 1 then
                    NextAvailableMessage = NextAvailableMessage - 1
                    Messages[z][MSG_MSG] = {}                
                end if
            else
                if Messages[z][MSG_Y] = CENTERED then
                    Why = (30 - (length(Messages[z][MSG_MSG]) / 2)) - 2
                    if Why < 1 then Why = 1 end if
                else Why = Messages[z][MSG_Y]
                end if
                position(Messages[z][MSG_X], Why)
                bk_color(0)
                if Messages[z][MSG_COLOR] = 0 then
                    text_color(rand(14) + 1)
                else text_color(Messages[z][MSG_COLOR])
                end if
                puts(1, 32 & Messages[z][MSG_MSG] & 32)
            end if
        end if
    end for
end procedure

procedure PlayWAV(sequence filename, atom wave_volume)
    object wave_file
    sequence channel_info
    
    if wave_volume = 0 then
        wave_volume = 100        
    end if
    
    if SoundOn then
        wave_file = load_wave(filename)
        if compare(wave_file, LOAD_WAVE_FAIL) != 0 then
            channel_info = play_wave(wave_file, 0, wave_volume)                        
        end if
    end if
end procedure

function ValidCoordinates(sequence c)
    if c[1] < 1 or c[2] < 1 or c[1] > 25 or c[2] > 60 then
        return FALSE
    else return TRUE
    end if
end function
             
procedure RunObjects(atom Board)
    atom z, k, TStart, CycleTime, alo, zz
    integer AdvanceF, RestartLoopThisF, RedrawF, LoopingBackF
    sequence kr, OldPos, C, VisLayers
    object dum1, dum2, dum3, dum4
    
    TStart = time()
    CycleTime = time()
    C = {}
    
    -- set to {0, 0, 0, 0} to not force redraw
    OldPos = repeat({0, 0, 0, 1}, length(Run))
    VisLayers = {1, 1, 1}
    
    if length(Run) != 0 then
        for ax = 1 to length(Run) do
            z = Run[ax]
        
            kr = get_keys()
            
            if length(kr) != 0 then
                k = kr[1]
            else k = -1
            end if
            
            if k = 1 then
                if Choice(17, "End this game?", {"Yes", "No"}, 1) = 1 then
                    exits = 2
                end if
            end if
        
            if Objects[z][1][OBJ_POS] != OBJ_HALTED then
                if Objects[z][1][OBJ_RUNSTATE] != 0 then
                    if CycleTime > (Objects[z][1][OBJ_RUNSTATE] +
                        (TPS * Objects[z][1][OBJ_CYCLE]))
                    then
                        Objects[z][1][OBJ_RUNSTATE] = 0                        
                    end if
                else
                    LoopingBackF = FALSE
                    
                    while TRUE do
                        AdvanceF = TRUE             -- advance object prog pos by def
                        RestartLoopThisF = FALSE    -- don't loop through again
                        RedrawF = FALSE             -- no full redraw
                        
                        -- C = current command being processed
                        -- pos is always ( OBJ_POS + 1 )
                    
                        if LoopingBackF = FALSE then
                            C = Objects[z][Objects[z][1][OBJ_POS] + 1]
                        end if
                        
                        if length(PreCommands[ax]) != 0 then
                            C = PreCommands[ax]
                            PreCommands[ax] = {}
                            AdvanceF = FALSE
                        end if
                        
                        OldPos[ax][1] = Objects[z][1][OBJ_X]
                        OldPos[ax][2] = Objects[z][1][OBJ_Y]
                        OldPos[ax][3] = Objects[z][1][OBJ_LAYER]
                        
                        dum3 = 0
                        
                        if length(C) != 0 then
                            -- Process level 1:
                            -- searching for variable inclusions
                            if atom(C[1]) then
                                if C[1] = 900 then
                                    if length(C) >= 2 then
                                        ScrollBuffers[ax] =
                                            append(ScrollBuffers[ax], C[2])
                                        dum3 = 999
                                    end if
                                end if
                            end if
                            
                            for a = 1 to length(C) do
                            
                                if atom(C[a]) then
                                if C[a] > 200 and C[a] < 300 then
                                
                                    if C[a] = 201 then  -- %me.x
                                        C[a] = num2str(Objects[z][1][OBJ_X])
                                    elsif C[a] = 202 then   -- %me.y
                                        C[a] = num2str(Objects[z][1][OBJ_Y])
                                    elsif C[a] = 203 then   -- %me.layer
                                        C[a] = num2str(Objects[z][1][OBJ_LAYER])
                                    elsif C[a] = 204 then   -- %curboard
                                        C[a] = num2str(Board)
                                    elsif C[a] = 205 then   -- %prop1
                                        C[a] = num2str(Objects[z][1][OBJ_PROPS + 0])
                                    elsif C[a] = 206 then   -- %prop2
                                        C[a] = num2str(Objects[z][1][OBJ_PROPS + 1])
                                    elsif C[a] = 207 then   -- %prop3
                                        C[a] = num2str(Objects[z][1][OBJ_PROPS + 2])
                                    elsif C[a] = 208 then   -- %prop4
                                        C[a] = num2str(Objects[z][1][OBJ_PROPS + 3])
                                    end if
                                end if
                                else
                                    -- finding flags
                                    if a = 2 then
                                        if C[1] = 15 then   -- #if
                                            if find(FlagNames, C[2]) then
                                                if FlagVals[find(FlagNames, C[2])] = TRUE then
                                                    C[2] = "1"
                                                else C[2] = {0}
                                                end if
                                            end if
                                        end if
                                    end if
                                
                                    dum1 = C[a]
                                    dum2 = 1
                                    if length(dum1) > 2 then   -- variable pull
                                        if dum1[1] = '%' then
                                            if a = 1 and length(C) > 1 then
                                                if C[a + 1] = 305 then
                                                    dum2 = 0
                                                end if
                                            end if
                                            if a != 1 then
                                                if C[1] = 38 or C[1] = 42 then   -- #variable, #addstat
                                                    dum2 = 0
                                                end if
                                            end if
                                            if dum2 = 1 then
                                                dum1 = dum1[2..length(dum1)]
                                                if find(dum1, VarNames) != 0 then
                                                    C[a] = num2str(VarVals[find(dum1, VarNames)])
                                                else C[a] = 0
                                                end if
                                            end if
                                        end if
                                    end if
                                end if
                            end for
                            
                            
                            -- searching for operators
                            alo = length(C) + 1
                            while TRUE do
                                alo = alo - 1
                                if alo = 0 then
                                    exit
                                end if
                                if atom(C[alo]) then
                                    if C[alo] = 209 then    -- alligned
                                        if alo != length(C) then
                                            zz = FindObject(C[alo + 1], Run)
                                            if zz != 0
                                            and zz != z then
                                                if Objects[zz][1][OBJ_X] = Objects[z][1][OBJ_X]
                                                or Objects[zz][1][OBJ_Y] = Objects[z][1][OBJ_Y] then
                                                    C[alo] = {'1'}
                                                else
                                                    C[alo] = {0}
                                                end if
                                            else C[alo] = {0}
                                            end if
                                            C = C[1..alo] & C[alo + 2..length(C)]
                                        end if
                                    -- Get the name of object at coordinates (p1,p2[,p3])
                                    elsif C[alo] = 210 then -- getobjname
                                        if length(C) > alo + 2 then
                                            if ValidCoordinates({str2num(C[alo + 1]), str2num(C[alo + 2])}) = TRUE then
                                                dum1 = ObjectPosTable[str2num(C[alo + 1])][str2num(C[alo + 2])][Objects[z][1][OBJ_LAYER]]
                                                if dum1 != 0 then
                                                    dum1 = Objects[dum1][1][OBJ_NAME]
                                                else dum1 = "nullobj"
                                                end if
                                                C = C[1..alo - 1] & {dum1} & C[alo + 3..length(C)]
                                            else C = C[1..alo - 1] & {"nullobj"} & C[alo + 3..length(C)]
                                            end if
                                        end if
                                    elsif C[alo] = 211 then -- keyb
                                        if alo != length(C) then                                            
                                            dum1 = C[alo + 1]
                                            dum2 = 0
                                            if compare(dum1, "up") = 0 then
                                                if k = KEY_UP then dum2 = 1 end if
                                            elsif compare(dum1, "down") = 0 then
                                                if k = KEY_DN then dum2 = 1 end if
                                            elsif compare(dum1, "left") = 0 then
                                                if k = KEY_LF then dum2 = 1 end if
                                            elsif compare(dum1, "right") = 0 then
                                                if k = KEY_RT then dum2 = 1 end if
                                            elsif compare(dum1, "space") = 0 then
                                                if k = 57 then dum2 = 1 end if
                                            elsif compare(dum1, "enter") = 0 then
                                                if k = 28 then dum2 = 1 end if
                                            end if
                                            if dum2 != 0 then
                                                C[alo] = "1"
                                            else C[alo] = {0}
                                            end if
                                            C = C[1..alo] & C[alo + 2..length(C)]
                                        end if
                                    elsif C[alo] = 212 then -- blocked
                                        dum3 = "1"
                                        if alo != length(C) then
                                            if atom(C[alo + 1]) then
                                                if C[alo + 1] = 220     -- north
                                                or C[alo + 1] = 221     -- south
                                                or C[alo + 1] = 222     -- east
                                                or C[alo + 1] = 223 then-- west
                                         
                                                    dum1 = C[alo + 1]       -- act dir
                                                    if dum1 = 220 then
                                                        dum1 = {-1, 0}
                                                    elsif dum1 = 221 then
                                                        dum1 = {1, 0}
                                                    elsif dum1 = 222 then
                                                        dum1 = {0, 1}
                                                    elsif dum1 = 223 then
                                                        dum1 = {0, -1}
                                                    end if
                                                    dum2 = {Objects[z][1][OBJ_X] + dum1[1], Objects[z][1][OBJ_Y] + dum1[2]}
                                                    if ValidCoordinates(dum2) = FALSE then
                                                        dum3 = "1"
                                                    else
                                                        if Boards[Board][BoardI(dum2[1], dum2[2], Objects[z][1][OBJ_LAYER], CHAR)] != 0
                                                        and Boards[Board][BoardI(dum2[1], dum2[2], Objects[z][1][OBJ_LAYER], NUM)] = 1 then
                                                            dum3 = "1"
                                                        else
                                                            if ObjectPosTable[dum2[1]][dum2[2]][Objects[z][1][OBJ_LAYER]] != 0 then
                                                                dum3 = "1"
                                                            else dum3 = {0}
                                                            end if
                                                        end if
                                                    end if
                                                end if
                                            end if
                                        end if
                                        C = C[1..alo - 1] & {dum3} & C[alo + 2..length(C)]
                                    elsif C[alo] = 213 then -- dop: seek
                                    elsif C[alo] = 214 then -- dop: rnd
                                        C[alo] = num2str(219 + rand(4))
                                    elsif C[alo] = 215 then -- dop: rndns
                                        C[alo] = num2str(219 + rand(2))
                                    elsif C[alo] = 216 then -- dop: rndew
                                        C[alo] = num2str(221 + rand(2))
                                    elsif C[alo] = 217 then -- dop: cw
                                    elsif C[alo] = 218 then -- dop: ccw
                                    elsif C[alo] = 219 then -- dop: opp
                                    elsif C[alo] = 301 then -- +
                                        if alo != length(C)
                                        and alo != 1 then                                            
                                            C[alo - 1] = num2str(
                                                str2num(C[alo - 1]) +
                                                str2num(C[alo + 1]))
                                            C = C[1..alo - 1] & C[alo + 2..length(C)]
                                        end if
                                        --alo = alo + 1
                                    elsif C[alo] = 302 then -- -
                                        if alo != length(C)
                                        and alo != 1 then
                                            C[alo - 1] = num2str(
                                                str2num(C[alo - 1]) -
                                                str2num(C[alo + 1]))
                                            C = C[1..alo - 1] & C[alo + 2..length(C)]
                                        end if
                                        --alo = alo - 1
                                    elsif C[alo] = 303 then -- *
                                        if alo != length(C)
                                        and alo != 1 then                                            
                                            C[alo - 1] = num2str(
                                                str2num(C[alo - 1]) *
                                                str2num(C[alo + 1]))
                                            C = C[1..alo - 1] & C[alo + 2..length(C)]
                                        end if
                                        --alo = alo + 1
                                    elsif C[alo] = 304 then -- /
                                        if alo != length(C)
                                        and alo != 1 then                                            
                                            C[alo - 1] = num2str(
                                                str2num(C[alo - 1]) /
                                                str2num(C[alo + 1]))
                                            C = C[1..alo - 1] & C[alo + 2..length(C)]
                                        end if
                                        --alo = alo + 1
                                    elsif C[alo] = 305 then -- =
                                        if alo != length(C)
                                        and alo != 1 then
                                        if sequence(C[alo - 1]) then
                                            dum1 = C[alo - 1]
                                            dum2 = 0
                                            if dum1[1] = '%' and alo = 2 then
                                                if length(dum1) > 1 then
                                                    -- assignment
                                                    while TRUE do
                                                        if dum1[1] = '%' then dum1 = dum1[2..length(dum1)]
                                                        else exit
                                                        end if
                                                    end while
                                                    if find(dum1, VarNames) = 0 then -- not defined 
                                                        VarNames = append(VarNames, dum1)
                                                        VarVals = append(VarVals, str2num(C[alo + 1]))
                                                    else
                                                        puts(1, dum1)
                                                        VarVals[find(dum1, VarNames)] = str2num(C[alo + 1])
                                                    end if
                                                    DrawGameBar(0, SideBar)
                                                end if
                                            else    -- value checking
                                                if sequence(C[alo - 1]) and sequence(C[alo + 1]) then
                                                    C[alo - 1] = num2str(str2num(C[alo - 1]) = str2num(C[alo + 1]))
                                                    if C[alo - 1][1] = 48 then C[alo - 1] = {0} end if
                                                    if length(C) > alo + 1 then
                                                        C = C[1..alo - 1] & C[alo + 2..length(C)]
                                                    else C = C[1..alo - 1]
                                                    end if
                                                end if
                                            end if
                                        end if
                                        end if
                                    end if
                                end if
                            end while

                            if atom(C[1]) then
                                if C[1] = 1 then    -- #_focus
                                    Objects[z][1][OBJ_FOCUS] = TRUE
                                elsif C[1] = 2 then -- #bind
                                    dum1 = FindObject(C[2], Run)
                                    if dum1 != 0 and dum1 != z then
                                        Objects[z] = Objects[z][1] & Objects[dum1][2..length(Objects[dum1])]
                                    end if
                                    Objects[z][1][OBJ_NAME] = Objects[dum1][1][OBJ_NAME]
                                    Objects[z][1][OBJ_POS] = 1
                                elsif C[1] = 3 then -- #change
                                elsif C[1] = 4 then -- #char
                                    if C[2][1] = 34 then
                                        Objects[z][1][OBJ_CHAR] = C[2][2]
                                    else
                                        Objects[z][1][OBJ_CHAR] = str2num(C[2])
                                    end if
                                elsif C[1] = 5 then -- #clear[flag]
                                    for a = 1 to length(FlagNames) do
                                        if compare(C[2], FlagNames[a]) = 0
                                            then FlagVals[a] = FALSE
                                            exit
                                        end if
                                    end for
                                elsif C[1] = 6 then -- #clone
                                    dum1 = C[2]
                                    dum2 = FindObject(dum1, Run)
                                    if dum2 != FALSE then
                                        Objects = append(Objects, Objects[dum2])
                                        Run = append(Run, length(Objects))
                                        if length(C) > 1 then   -- x, y
                                            Objects[length(Objects)][OBJ_X] = str2num(C[3])
                                            Objects[length(Objects)][OBJ_Y] = str2num(C[4])
                                            if length(C) > 3 then   -- layer
                                                Objects[length(Objects)][OBJ_LAYER] = str2num(C[5])                                                
                                            end if
                                        end if
                                    end if
                                elsif C[1] = 7 then -- #color
                                    if C[2][1] = 'c' then   -- MZX hex-color value
                                        dum1 = C[2][2]
                                        if length(C[2]) > 2 then
                                            dum2 = C[2][2]
                                        else dum2 = '0'
                                        end if
                                        dum1 = hexchar_to_atom(dum1)
                                        dum2 = hexchar_to_atom(dum2)
                                        Objects[z][1][OBJ_COLOR] = ColAttrib(dum1, dum2)
                                    else    -- ff bb color
                                        Objects[z][1][OBJ_COLOR] = ColAttrib(str2num(C[2]), str2num(C[3]))
                                    end if
                                elsif C[1] = 8 then -- #cycle
                                    Objects[z][1][OBJ_CYCLE] = str2num(C[2])
                                elsif C[1] = 9 then -- #die
                                    DeleteObject(z, Board)
                                    Run = BuildRunTable(Board)
                                    return
                                elsif C[1] = 10 then -- #end
                                    Objects[z][1][OBJ_POS] = OBJ_HALTED
                                elsif C[1] = 11 then -- #endgame
                                    exits = 2
                                elsif C[1] = 12 then -- #give
                                elsif C[1] = 13 then -- #go
                                elsif C[1] = 14 then -- #idle
                                    Objects[z][1][OBJ_RUNSTATE] = CycleTime
                                elsif C[1] = 15 then -- #if
                                    if sequence(C[2]) then
                                        for a = 1 to length(FlagNames) do
                                            if compare(lower(C[2]), lower(FlagNames[a])) = 0 then
                                                if FlagVals[a] then
                                                    C[2] = "1"
                                                else C[2] = {0}
                                                end if
                                                exit
                                            end if
                                        end for
                                    end if
                                    if str2num(C[2]) != 0 then
                                        dum1 = C
                                        C = {}
                                        for a = 4 to length(dum1) do
                                            if atom(dum1[a]) then
                                                if dum1[a] = 401 then -- else
                                                    exit
                                                end if
                                            end if
                                            C = append(C, dum1[a])
                                        end for
                                        k = -1
                                        RestartLoopThisF = TRUE
                                    else
                                        for a = 4 to length(C) do
                                            if atom(C[a]) then
                                                if C[a] = 401 then  -- else
                                                    dum1 = C
                                                    C = {}
                                                    for b = a + 1 to length(dum1) do
                                                        C = append(C, dum1[b])
                                                    end for
                                                    k = -1
                                                    RestartLoopThisF = TRUE
                                                    exit
                                                end if
                                            end if
                                        end for
                                    end if
                                elsif C[1] = 16 then -- #loadfont
                                elsif C[1] = 17 then -- #loadmod
                                    if SoundOn then
                                        dum2 = 0
                                        if length(C) >= 2 then
                                            if sequence(C[2]) then
                                                dum1 = C[2]
                                                if load_mod(dum1) then
                                                    dum2 = 1
                                                end if
                                            else dum2 = 1
                                            end if
                                        else dum2 = 1
                                        end if
                                        if dum2 != 0 then
                                            GameError({"Could not load MOD file:", C[2]}, "Error!")
                                        else unpause_sound()
                                        end if
                                    end if
                                elsif C[1] = 18 then -- #playmod
                                    if SoundOn then
                                        unpause_sound()
                                    end if
                                elsif C[1] = 19 then -- #pausemod
                                    if SoundOn then
                                        pause_sound()
                                    end if
                                elsif C[1] = 20 then -- #move
                                    if length(C) = 3 then   -- x, y
                                        dum1 = str2num(C[2])
                                        dum2 = str2num(C[3])
                                        if (dum1 > 0 and dum1 <= 25)
                                        and (dum2 > 0 and dum2 <= 60) then
                                            Objects[z][1][OBJ_X] = dum1
                                            Objects[z][1][OBJ_X] = dum2
                                        end if
                                    end if
                                    if length(C) = 4 then   -- x, y, layer
                                        dum1 = str2num(C[2])
                                        dum2 = str2num(C[3])
                                        dum3 = str2num(C[4])
                                        if (dum1 > 0 and dum1 <= 25)
                                        and (dum2 > 0 and dum2 <= 60)
                                        and (dum3 > 0 and dum3 <= 3) then
                                            Objects[z][1][OBJ_X] = dum1
                                            Objects[z][1][OBJ_X] = dum2
                                            Objects[z][1][OBJ_LAYER] = dum3
                                        end if
                                    end if
                                elsif C[1] = 21 then -- #playwav
                                    PlayWAV(C[2], 100)
                                elsif C[1] = 22 then -- #lock
                                    Objects[z][OBJ_LOCKED] = TRUE
                                elsif C[1] = 23 then -- #redraw
                                    DrawBoard(CurBoard, VisibleLayers)
                                    DrawObjectsInGame(Run, repeat({0,0,0,1}, length(Run)), CurBoard)
                                elsif C[1] = 24 then -- #restart
                                    Objects[z][OBJ_POS] = 1
                                elsif C[1] = 25 then -- #restore
                                    dum2 = trim(lower(C[2]))
                                    dum2 = ":|" & dum2
                                    for a = 1 to length(Objects[z]) do
                                        if compare(lower(Objects[z][a]),
                                                dum2) = 0 then
                                            Objects[z][a] = ":" &
                                            Objects[z][a][3..length(Objects[z][a])]
                                            exit
                                        end if
                                    end for
                                elsif C[1] = 26 then -- #scroll
                                elsif C[1] = 27 then -- #send
                                    dum2 = lower(C[2][find(':',
                                                C[2]) + 1..length(
                                                C[2])])
                                            
                                    dum1 = {}
                                    if find(':', C[2]) then
                                        dum1 = trim(lower(C[2][1..find(':',
                                                        C[2]) - 1]))
                                    elsif length(C) > 2 then
                                        if C[3][1] = ':' then
                                            dum1 = trim(lower(C[2]))
                                            dum2 = trim(lower(C[3][2..length(C[3])]))
                                        end if
                                    end if
                                    
                                    alo = -1
                                    if length(dum1) != 0 then
                                        for aa = 1 to length(Run) do
                                            if compare(lower(Objects[Run[aa]][1]
                                                        [OBJ_NAME]), dum1) = 0 then
                                                alo = Run[aa]
                                                if Objects[alo][1][OBJ_LOCKED] = FALSE then
                                                    for a = 2 to length(Objects[alo]) do
                                                        if sequence(Objects[alo][a][1]) then
                                                        if compare(lower(Objects[alo][a][1]
                                                                [2..length(Objects[alo][a][1])]),
                                                                dum2) = 0 then
                                                            Objects[alo][1][OBJ_POS] = a
                                                            Objects[alo][1][OBJ_RUNSTATE] = 1
                                                            if alo = z then AdvanceF = FALSE end if
                                                            exit
                                                        end if
                                                        end if
                                                    end for
                                                end if
                                            end if
                                        end for
                                    else
                                        alo = z
                                        if Objects[alo][1][OBJ_LOCKED] = FALSE then
                                            for a = 2 to length(Objects[alo]) do
                                                if sequence(Objects[alo][a][1]) then
                                                if compare(lower(Objects[alo][a][1]
                                                        [2..length(Objects[alo][a][1])]),
                                                        dum2) = 0 then
                                                    Objects[alo][1][OBJ_POS] = a
                                                    Objects[alo][1][OBJ_RUNSTATE] = 1
                                                    if alo = z then AdvanceF = FALSE end if
                                                    exit
                                                end if
                                                end if
                                            end for
                                        end if
                                    end if
                                elsif C[1] = 28 then -- #set[flag]
                                    dum1 = FALSE
                                    for a = 1 to length(FlagNames) do
                                        if compare(C[2], FlagNames[a]) = 0
                                            then FlagVals[a] = TRUE
                                            dum1 = TRUE
                                            exit
                                        end if
                                    end for
                                    if dum1 = FALSE then
                                        FlagNames = append(FlagNames, C[2])
                                        FlagVals = append(FlagVals, TRUE)
                                    end if
                                elsif C[1] = 29 then -- #set_msx
                                    if length(C) >= 2 then
                                        MessageX = str2num(C[2])
                                    end if
                                elsif C[1] = 30 then -- #set_msy
                                    if length(C) >= 2 then
                                        MessageY = str2num(C[2])
                                    end if
                                elsif C[1] = 31 then -- #set_mscolor
                                    if length(C) >= 2 then
                                        MessageColor = str2num(C[2])
                                    end if
                                elsif C[1] = 32 then -- #set_mstime
                                    if length(C) >= 2 then
                                        MessageTime = str2num(C[2])
                                    end if
                                elsif C[1] = 33 then -- #take
                                elsif C[1] = 34 then -- #transport
                                    if compare(C[2], num2str(str2num(C[2]))) != 0 then
                                        dum1 = {}
                                        for a = 2 to length(C) do
                                            dum1 = dum1 & (C[a] & ' ')
                                        end for
                                        dum1 = dum1[1..length(dum1) - 1]
                                        for a = 1 to length(Boards) do
                                            if compare(dum1, BoardProps[a][BP_NAME]) = 0 then
                                                dum1 = a
                                                exit
                                            end if
                                        end for
                                    else dum1 = str2num(C[2])
                                    end if
                                    if atom(dum1) then
                                        CurBoard = dum1
                                        DrawGameBar(0, SideBar)
                                        InitWorld_2()
                                        if AdvanceF = TRUE then
                                            Objects[z][1][OBJ_POS] = Objects[z][1][OBJ_POS] + 1
                                        end if
                        
                                        if Objects[z][1][OBJ_POS] > (length(Objects[z]) - 1) then
                                            Objects[z][1][OBJ_POS] = OBJ_HALTED
                                        end if
                                        return
                                    end if
                                elsif C[1] = 35 then -- #unlock
                                    Objects[z][OBJ_LOCKED] = FALSE
                                elsif C[1] = 36 then -- #vis[ible]layers
                                elsif C[1] = 37 then -- #zap
                                    dum2 = trim(lower(C[2]))
                                    dum2 = ":" & dum2
                                    for a = 1 to length(Objects[z]) do
                                        if compare(lower(Objects[z][a]),
                                                dum2) = 0 then
                                            Objects[z][a] = ":|" &
                                            Objects[z][a][2..length(Objects[z][a])]
                                            exit
                                        end if
                                    end for
                                elsif C[1] = 38 then -- #variable
                                    if length(C[2]) != 1 then
                                        while TRUE do
                                            if C[2][1] = '%' then
                                                C[2] = C[2][2..length(C[2])]
                                            else exit
                                            end if
                                        end while
                                        if length(C[2]) != 0 then
                                            VarNames = append(VarNames, C[2])
                                            VarVals = append(VarVals, 0)
                                        end if
                                    end if
                                elsif C[1] = 39 then -- #transportnof
                                elsif C[1] = 40 then -- #scrollarea
                                elsif C[1] = 41 then -- #changearea
                                elsif C[1] = 42 then -- #addstat
                                    if length(C) > 1 then
                                        if length(C) > 2 then
                                            if length(C) > 3 then
                                                -- item w/ title, char, var
                                                C[4] = C[4][2..length(C[4])]
                                                SideBar = append(SideBar, {2, C[2], C[3], C[4]})
                                                if find(C[4], VarNames) = 0 then -- not defined
                                                    VarNames = append(VarNames, C[4])
                                                    VarVals = append(VarVals, 0)
                                                else
                                                    VarVals[find(C[4], VarNames)] = 0
                                                end if
                                            else
                                                -- item w/ title, var
                                                C[3] = C[3][2..length(C[3])]
                                                SideBar = append(SideBar, {2, C[2], 0, C[3]})
                                                if find(C[3], VarNames) = 0 then -- not defined 
                                                    VarNames = append(VarNames, C[3])
                                                    VarVals = append(VarVals, 0)
                                                else
                                                    VarVals[find(C[3], VarNames)] = 0
                                                end if
                                            end if
                                        else
                                            -- title
                                            SideBar = append(SideBar, {2, {}, 0, C[2]})
                                        end if
                                        DrawGameBar(0, SideBar)
                                    end if
                                elsif C[1] = 43 then -- #delstat
                                    SideBar = SideBar[1..length(SideBar) - 1]
                                    DrawGameBar(0, SideBar)
                                elsif C[1] = 44 then -- #volume
                                elsif C[1] = 45 then -- #volup
                                    volume_up()
                                elsif C[1] = 46 then -- #voldown
                                    volume_down()
                                elsif C[1] = 101 or C[1] = 106  -- /n, \n
                                or C[1] = 102 or C[1] = 107     -- /s, \s
                                or C[1] = 103 or C[1] = 108     -- /e, \e
                                or C[1] = 104 or C[1] = 109 then-- /w, \w
                                    -- store old position
                                    dum1 = {Objects[z][1][OBJ_X], Objects[z][1][OBJ_Y]}
                                    dum2 = C[1]     -- movement dir
                                    dum3 = 1        -- auto /i flag
                                    if dum2 >= 106 then -- \n\s\e\w command - no /i
                                        dum2 = dum2 - 5 -- change to std /n/s/e/w
                                        dum3 = 0    -- set to no /i
                                    end if
                                    dum2 = dum2 - 100   -- 1,2,3,4 - /n/s/e/w now
                                    
                                    -- dest: {constf, setx, sety}
                                    dum4 = {}                                    
                                    if dum2 = 1 then    -- n
                                        Objects[z][1][OBJ_X] = Objects[z][1][OBJ_X] - 1
                                        if Objects[z][1][OBJ_X] < 1 then dum4 = {BP_BN, 25, -1} end if
                                    elsif dum2 = 2 then -- s
                                        Objects[z][1][OBJ_X] = Objects[z][1][OBJ_X] + 1
                                        if Objects[z][1][OBJ_X] > 25 then dum4 = {BP_BS, 1, -1} end if
                                    elsif dum2 = 3 then -- e
                                        Objects[z][1][OBJ_Y] = Objects[z][1][OBJ_Y] + 1
                                        if Objects[z][1][OBJ_Y] > 60 then dum4 = {BP_BE, -1, 1} end if
                                    elsif dum2 = 4 then -- w
                                        Objects[z][1][OBJ_Y] = Objects[z][1][OBJ_Y] - 1
                                        if Objects[z][1][OBJ_Y] < 1 then dum4 = {BP_BW, -1, 60} end if
                                    end if
                                    
                                    if length(dum4) != 0 then    -- off board
                                        if length(BoardProps[Board][dum4[1]]) != 0 then
                                            if dum4[1] != -1 then   -- x
                                                if Boards[BoardProps[Board][BP_BN][1]][BoardI(dum4[2], Objects[z][1][OBJ_Y], Objects[z][1][OBJ_LAYER], CHAR)] = 0
                                                or Boards[BoardProps[Board][BP_BN][1]][BoardI(dum4[2], Objects[z][1][OBJ_Y], Objects[z][1][OBJ_LAYER], CHAR)] = 32 then
                                                    -- move it to this board
                                                    Objects[z][1][OBJ_X] = dum4[2]
                                                    Objects[z][1][OBJ_BOARD_NUM] = BoardProps[Board][dum4[1]][1]
                                                    Run = BuildRunTable(Board)
                                                    BuildObjectTable(Board)
                                                    DrawBoard(Board, VisLayers)
                                                else
                                                    -- blocked
                                                    Objects[z][1][OBJ_X] = dum1[1]
                                                    Objects[z][1][OBJ_Y] = dum1[2]
                                                end if
                                            else
                                                if Boards[BoardProps[Board][BP_BN][1]][BoardI(Objects[z][1][OBJ_X], dum4[3], Objects[z][1][OBJ_LAYER], CHAR)] = 0
                                                or Boards[BoardProps[Board][BP_BN][1]][BoardI(Objects[z][1][OBJ_X], dum4[3], Objects[z][1][OBJ_LAYER], CHAR)] = 32 then
                                                    -- move it to this board
                                                    Objects[z][1][OBJ_Y] = dum4[3]
                                                    Objects[z][1][OBJ_BOARD_NUM] = BoardProps[Board][dum4[1]][1]
                                                    Run = BuildRunTable(Board)
                                                    BuildObjectTable(Board)
                                                    DrawBoard(Board, VisLayers)
                                                else
                                                    -- blocked
                                                    Objects[z][1][OBJ_X] = dum1[1]
                                                    Objects[z][1][OBJ_Y] = dum1[2]
                                                end if
                                            end if
                                        else
                                            Objects[z][1][OBJ_X] = dum1[1]
                                            Objects[z][1][OBJ_Y] = dum1[2]
                                        end if
                                    end if
                                    
                                    if Objects[z][1][OBJ_X] != dum1[1]
                                    or Objects[z][1][OBJ_Y] != dum1[2] then
                                        UpdateObjectTable(z, Board, 0)
                                        DrawObjectsInGame(Run, OldPos, Board)
                                        Objects[z][1][OBJ_RUNSTATE] = CycleTime
                                    end if
                                elsif C[1] = 105 then -- /i
                                    Objects[z][1][OBJ_RUNSTATE] = CycleTime
                                end if
                            else
                                if C[1][1] = '@' then   -- name assignment
                                    Objects[z][1][OBJ_NAME] = C[1][2..length(C[1])]
                                elsif C[1][1] = ':'     -- label - skip
                                or C[1][1] = '|' then   --
                                    -- future stuff... maybe
                                elsif C[1][1] != '%' then
                                    if atom(C[1]) then
                                        if C[1] != 0 then
                                            GameError({"Bad command:", CompileData[C[1]]}, "Error!")
                                        end if
                                    else
                                        if C[1][1] != 0 then
                                            GameError({"Bad/unsupported command:", C[1]}, "Error!")
                                        end if
                                    end if
                                end if
                            end if
                        end if
                        
                        if atom(dum3) then
                            if dum3 != 999 then
                                if length(ScrollBuffers[ax]) = 1 then
                                    NewMessage({ScrollBuffers[ax][1],
                                                MessageColor,
                                                MessageTime,
                                                MessageX,
                                                MessageY})
                                    ScrollBuffers[ax] = {}
                                    UpdateMessages()
                                elsif length(ScrollBuffers[ax]) > 1 then
                                    dum1 =
                                    DisplayScroll(Objects[Run[ax]][1][OBJ_NAME], ScrollBuffers[ax], 1)
                                    ScrollBuffers[ax] = {}
                                    if length(dum1) != 0 then
                                        PreCommands[ax] = {27, dum1}
                                    end if
                                end if
                            end if
                        end if
                        
                        if AdvanceF = TRUE then
                            Objects[z][1][OBJ_POS] = Objects[z][1][OBJ_POS] + 1
                        end if
                        
                        if Objects[z][1][OBJ_POS] > (length(Objects[z]) - 1) then
                            Objects[z][1][OBJ_POS] = OBJ_HALTED
                        end if
                          
                        if RestartLoopThisF = TRUE then
                            LoopingBackF = TRUE
                        else exit
                        end if
                    end while
                end if
            end if
        end for
    else
        if CurBoard != WorldProps[WP_TBOARD][1] then
            -- no objects, and this is only OK for the title board
            GameError("Board " & num2str(CurBoard) & " has no objects!", "Error!")
            exits = 1
        end if
    end if
    
    if CurBoard = WorldProps[WP_TBOARD][1] then
            k = upper(get_key())
            clear_keys()
            if k = KEY_ESC or k = 'Q' then
                if Choice(17, "Quit?", {"Yes", "No"}, 1) = 1 then
                    exits = 1
                end if
            elsif k = 'W' then
                SelectWorld(0)
                DrawGameBar(1, {})
                InitWorld_2()
            elsif k = 'P' then
                CurBoard = WorldProps[WP_SBOARD][1]
                DrawGameBar(0, {})
                InitWorld_2()
            elsif k = 'L' then
                dum1 = DisplayScroll("Not available", {{}, "Load/save options are not implemented yet.",{}}, 2)
            elsif k = 'A' then
                dum1 = DisplayScroll("About ZIG",
                            {"$- - - -",{},
                             "$ZIG",{},
                             "$- - - -",{},
                             "$Version 1.0",
                             "$Copyright 1997-99, Jacob Hammond.",{},{},
                             "ZIG has been and is a work in progress; it",
                             "was begun in March of 1998. It was inspired",
                             "by Epic [Mega]Games' classic game ZZT, thus",
                             "the name ZZT-Inspired Game Creation System.",
                             "ZIG is not a clone of ZZT, more of a",
                             "remake of sorts.",{},
                             "$Webpage",
                             "           http://surf.to/zig",{},
                             "$or",
                             "  http://www2.crosswinds.net/~zig/index.htm",{},
                             "$E-mail",
                             "            zig@crosswinds.net",{},{},
                             "$Special thanks go out to:",{},
                             "$xabbott",
                             "For being interested in ZIG from the start,",
                             "for doing a special feature on ZIG on",
                             "Meltdown, and for general support and ideas.",{},
                             "$Spock (I think your name may have changed)",
                             "For many ideas over the course of ZIG's devel-",
                             "opment. (How about some more? :-)",{},
                             "$wep",
                             "For interest and support for ZIG. ZZT++ is",
                             "going to be good!", {},
                             "$Bluenatic",
                             "For all of your detailed bug reports.",{},
                             "$Hydra78",
                             "For continued interest and bug reports.",{},
                             "$Jeremy Brannon",
                             "For being interested in ZIG from the very start.",{},
                             "$Gunner",
                             "For *lots* of ideas about multiplayer support in",
                             "ZIG.",{},
                             "$Greetings (in no order) to:",{},
                             "Chronos30",
                             "Gunner",
                             "Frank Green",
                             "Flatcoat Lab",
                             "Dark FLR",
                             "Daemon Lotos",
                             "MZT",
                             "Hercules",
                             "Ethan Hunt",
                             "Akwende",
                             "Everyone who beta-tested",
                             "Everyone in the ZZT community",
                             {}
                             }, 5)
            end if
    end if
end procedure

atom mup

exits = 0

Init()

SelectWorld(1)
InitWorld()


while exits != 1 do
    CurBoard = WorldProps[WP_TBOARD][1]
    DrawGameBar(1, {})
    InitWorld_2()
    
    mup = time() + 1

    while exits = 0 do
        RunObjects(CurBoard)
        if time() > mup then
            mup = time() + .1
            UpdateMessages()
        end if
    end while
    
    if exits = 2 then exits = 0 end if
end while

-- DISABLED because it slows it up
--fade_out(2, FALSE)  --FALSE for now, goofs up stereo
EndProg(0)

