-- zsidebar.e
-- ZIG sidebar display routines

include zgfx.e
include zutil.e

global constant
    -- Sidebar button hilight (SB_HLITE1 for c03, SB_HLITE2 for c07)
    SB_HLITE1 = 16, SB_HLITE2 = 17,

    -- Constants for SB_Input()
    SBI_FILENAME = 1, SBI_NUMBER = 2, SBI_ANYSTR = 3


global procedure SB_Clear()
    display_text_image({1, 61}, repeat({32,31,32,31,32,31,32,31,32,31,32,31,32,31,32,31,32,31,32,31,32,31,32,31,32,31,32,31,32,31,32,31,32,31,32,31,32,31,32,31}, 25))
end procedure

global procedure SB_Text(atom X, atom C, sequence Text)
    if C = SB_HLITE1 then
        text_color(0) bk_color(3)
    elsif C = SB_HLITE2 then
        text_color(0) bk_color(7)
    else
        text_color(C) bk_color(1)
    end if
    
    position(X, 62)
    puts(1, Text)
end procedure

global procedure SB_CText(atom X, atom C, sequence Text)
    if C = SB_HLITE1 then
        text_color(0) bk_color(3)
    elsif C = SB_HLITE2 then
        text_color(0) bk_color(7)
    else
        text_color(C) bk_color(1)
    end if
    
    position(X, 62 + (9 - (length(Text) / 2)))
    puts(1, Text)
    
end procedure

global procedure SB_BtnText(atom X, atom C, sequence Text)
    SB_Text(X, 14, Text)
    
    text_color(0)
    bk_color(C)
    position(X, 62)
    if Text[3] = 32 then
        puts(1, Text[1..3])
    else puts(1, Text[1..4])
    end if
end procedure

global procedure SB_M_DrawCTile(atom Pos, atom Old)
    text_color(15)
    bk_color(1)
    
    if Old < 18 then
        position(4, 61 + Old)
        puts(1, ' ')
    else
        position(13, 79)
        puts(1, ' ')
    end if

    if Pos < 18 then
        position(4, 61 + Pos)
        puts(1, '')
    else
        position(13, 79)
        puts(1, '')    
    end if
end procedure

global procedure SB_M_DrawFWP(atom W1)
    position(2, 72)
    text_color(14)
    bk_color(1)
    if W1 then
        puts(1, "*Floors")
    else puts(1, "*Walls ")
    end if
end procedure

global procedure SB_M_DrawCol(atom C)
    display_text_image({8, 73}, {{254, C}})
end procedure        

global procedure SB_M_DrawLayer(atom L)
    position(17, 74)
    text_color(15)
    bk_color(1)
    ?L
    text_color(9)
    position(17, 75)
    puts(1, "/3")
    text_color(15)
end procedure

global procedure SB_M_DrawCustTs()
    position(3, 62)
        
    text_color(15)
    bk_color(0)
     
    for z = 1 to length(CustTiles) do
        if CustTiles[z] != 0 then
            display_text_image({3, 61 + z}, {{CustTiles[z], 15}})        
        else display_text_image({3, 61 + z}, {{'C', 8}})
        end if
    end for    
end procedure

global procedure SB_M_DrawDrawStat(atom s)
    position(21, 62)
    text_color(15)
    bk_color(1)
    
    if s = 0 then
        puts(1, "Drawing Off")
    elsif s = 1 then
        puts(1, "Drawing On ")
    elsif s = 2 then
        puts(1, "Typing Mode")        
    end if
end procedure

global procedure SB_M_DrawShowns(sequence S)
    position(18, 73)
    bk_color(1)
    for z = 1 to 3 do
        if S[z] then text_color(15) else text_color(9) end if
        printf(1, "%1d", z)
    end for
end procedure

global procedure SB_M_DrawGrabbedObj(sequence Obj)
    display_text_image({14, 79},
        {{Obj[1][OBJ_CHAR], Obj[1][OBJ_COLOR]}})
end procedure

global procedure SB_M_DrawCBoard(atom b)
    position(19, 73)
    text_color(14)
    bk_color(1)
    printf(1, "%3d", b)
end procedure

global procedure SB_M_Draw()
    SB_Clear()
    SB_CText(1, SB_HLITE1, " - ZIG Editor - ")

    Box(2, 61, 9, 79, 1, "Patterns")
        
    SB_M_DrawCustTs()    
    SB_M_DrawCTile(1, 1)
    SB_M_DrawFWP(0)
    SB_M_DrawCol(15)
    SB_M_DrawLayer(2)
    SB_M_DrawDrawStat(0)
    SB_M_DrawShowns({1, 1, 1})
    SB_M_DrawGrabbedObj(GrabbedObject)
    SB_M_DrawCBoard(1)
    
    SB_BtnText(5, 3, " P  Select Pat.")
    SB_BtnText(6, 7, " Z  Floor/Wall") 
    SB_BtnText(7, 3, " D  Choose Char")
    SB_BtnText(8, 7, " C  Color:")

    SB_BtnText(10, 3, " N  New World")
    SB_BtnText(11, 7, " L  Load World")
    SB_BtnText(12, 3, " S  Save World")
    
    SB_BtnText(14, 7, " O  Place Object")
    SB_BtnText(15, 3, " F1  Libraries")
    SB_BtnText(16, 7, " F2  Typing Mode")
    SB_BtnText(17, 3, " `~  Layer: ")
    SB_BtnText(18, 7, "1..3 Vis.: ")
    SB_BtnText(19, 3, " B  Board: ")
        
end procedure

global function SB_Input(atom TypeOf, atom Line, sequence Prompt, atom Max)
    atom Pos, Key
    sequence Ext, Inp, SS
    
    SS = SaveScreen()

    Inp = {}
    Pos = 1
    
    if TypeOf = SBI_FILENAME then
        Ext = "." & Prompt[length(Prompt) - 2..length(Prompt)]
        Prompt = Prompt[1..length(Prompt) - 3]
        Max = 8
    else Ext = {}
    end if
    
    if TypeOf = SBI_ANYSTR and Max > 18 then Max = 18 end if
    
    position(Line, 62)
    text_color(14)
    bk_color(1)
    puts(1, Prompt)
    position(Line + 1, 62)
    text_color(15)
    bk_color(0)
    puts(1, repeat(32, Max))
    
    if TypeOf = SBI_FILENAME then
        position(Line + 1, 70)
        puts(1, Ext)
    end if
    
    cursor(UNDERLINE_CURSOR)
    position(Line + 1, 62)
    Key = -1
    
    while Key != KEY_ESC and Key != KEY_CR do
        Key = get_key()
        clear_keys()
        if TypeOf = SBI_FILENAME then
            if upper(Key) != Key or lower(Key) != Key then
                if length(Inp) < Max then
                    Key = upper(Key)
                    Inp = append(Inp, Key)
                    position(Line + 1, 62)
                    puts(1, Inp)
                end if
            end if                
        else
            if Key > 31 and Key < 127 and length(Inp) < Max then                
                Inp = append(Inp, Key)
                position(Line + 1, 62)
                puts(1, Inp)
            end if
        end if
        if Key = 8 then
            if length(Inp) != 0 then
                Inp = Inp[1..length(Inp) - 1]
                position(Line + 1, 62)
                puts(1, Inp & " ")
                position(Line + 1, 62)
                puts(1, Inp)
            end if
        end if        
    end while
    
    cursor(NO_CURSOR)
    ResScreen(SS)

    return Inp
end function

global procedure NOPut(atom NO)
    SB_Text(24, 14, "Objects: ")
    printf(1, "%3d", NO)
end procedure

global function Choice(atom x, sequence prompt, sequence choices, atom selected)
    atom default, k, os
    sequence ss

    ss = SaveScreen()
    default = selected
    selected = -1
    
    SB_Text(x, 14, prompt)
    
    while TRUE do
        os = selected
        k = get_key()
        clear_keys()
        
        if selected = -1 then
            selected = default
        end if
        
        if k = KEY_CR then
            ResScreen(ss)
            return(selected)
        elsif k = KEY_ESC then
            ResScreen(ss)
            return(default)
        elsif k = KEY_LF then
            if selected != 1 then
                selected = selected - 1
            end if
        elsif k = KEY_RT then
            if selected != length(choices) then
                selected = selected + 1
            end if
        end if

        if selected != os then
            position(x + 1, 62)
            for z = 1 to length(choices) do
            if z = selected then
                text_color(0)
                bk_color(3)
            else
                text_color(15)
                bk_color(1)
            end if 
            puts(1, " " & choices[z] & " ")
            end for
        end if
    end while
end function


