include misc.e
include wildcard.e
include file.e
include get.e

--include zchar.e
--include palette.e
--include zutil.e
--include zgfx.e
--include zsidefnc.e

global constant FONT_WIDTH = 8, FONT_HEIGHT = 16
global sequence WholeSet  --, CharSet     -- remove charset for inclusion in main

egachars()
toggle_blinking(0)

global procedure EditChars_drawchar(sequence pat)
    sequence d, e
    
    d = {}
    for z = 1 to 14 do
        e = {}
        for a = 1 to FONT_WIDTH do
            if pat[z][a] = 'x' then
                e = e & "ۇۇ"
            else e = e & ""
            end if
        end for
        d = append(d, e)
    end for
    
    display_text_image({5, 6}, d)
end procedure

function EditChars_makepattern(atom curchar)
    sequence bits, ret

    ret = repeat(repeat(0, FONT_WIDTH), FONT_HEIGHT)
    for z = 1 to FONT_HEIGHT do
        bits = int_to_bits(CharSet[curchar + 1][z], 8)
        for a = 1 to 8 do
            if bits[a] = 1 then
                ret[z][9 - a] = 'x'
            end if
        end for
    end for
    
    return ret
end function

global procedure EditChars()
    atom curchar, x, y, draw, k, ox, oy, ct, t, mod_flag, od, f
    sequence cc, fn, bufcc, copy_buf, SS

    curchar = 0
    x = 1
    y = 1
    draw = 0
    t = 0
    copy_buf = {}

    SS = SaveScreen()    
    Box(4, 4, 24, 56, 1, "Character Set Editor")
    
    cc = EditChars_makepattern(curchar)
    EditChars_drawchar(cc)
    
    text_at(6, 25, "Current character: " & sprintf("%3d", curchar), 15)
    display_text_image({6, 50}, {{curchar, 15}})
    text_at(7, 25, "+/- - select character to edit", 7)
    text_at(8, 25, "Arrows - move cursor in ed.box", 7)
    text_at(9, 25, "Alt+arrows - select character", 7)
    text_at(10, 25, "F1/F2 - copy/paste character", 7)
    text_at(11, 25, "Space - set/clear current pixel", 7)
    text_at(12, 25, "Tab - toggle draw mode", 7)
    text_at(13, 25, "Ctrl+arrows - shift character", 7)
    text_at(14, 25, "H/V - flip horiz./vert.", 7)
    text_at(15, 25, "C - clear current character", 7)
    text_at(16, 25, "S - save current set", 7)
    text_at(17, 25, "L - load set", 7)
    
    display_text_image({16, 23}, WholeSet)

    ct = 0
    add_all_fonts(0, CharSet)
    display_text_image({4 + x, 5 + y}, {"" & 14 & "" & 14})
                                 
    while 1 do
        k = get_key()
        ox = x
        oy = y
        od = draw
        mod_flag = 0
        
        if k = KEY_ESC then
            ResScreen(SS)
            return
        elsif k = KEY_UP then
            x = x - 1
        elsif k = KEY_DN then
            x = x + 1
        elsif k = KEY_LF then
            y = y - 1
        elsif k = KEY_RT then
            y = y + 1
        elsif k = KEY_F1 then
            copy_buf = cc
        elsif k = KEY_F2 then
            if length(copy_buf) != 0 then
                cc = copy_buf
                CharSet[curchar + 1] = pattern_to_binary(cc)
                EditChars_drawchar(cc)
                mod_flag = 1
            end if
        elsif k = 'C' or k = 'c' then
            cc = EditChars_makepattern(0)
            CharSet[curchar + 1] = pattern_to_binary(cc)
            EditChars_drawchar(cc)
            mod_flag = 1
        elsif k = ' ' then
            if cc[x][y] = 0 then
                cc[x][y] = 'x'
            else cc[x][y] = 0
            end if
            mod_flag = 1
        elsif k = 'S' or k = 's' then
            fn = SB_Input(SBI_FILENAME, 20, "Filename?.ZCH", 8)
            fn = fn & ".zch"
            f = open(fn, "wb")
            for z = 1 to 256 do
                for a = 1 to FONT_HEIGHT do
                    puts(f, CharSet[z][a])
                end for
            end for
            close(f)
        elsif k = 'L' or k = 'l' then
            fn = SB_Input(SBI_FILENAME, 20, "Filename?.ZCH", 8)
            fn = fn & ".zch"
            ZIGLoadFont(fn)
        elsif k = KEY_TAB then
            if draw = 0 then draw = 1 else draw = 0 end if
        elsif k = '+' or k = '=' or k = KEY_PGUP or k = KEY_ALT_RT then
            if curchar != 255 then
                CharSet[curchar + 1] = pattern_to_binary(cc)
                curchar = curchar + 1
                cc = EditChars_makepattern(curchar)
                EditChars_drawchar(cc)
                bk_color(1)
                text_at(6, 25, "Current character: " & sprintf("%3d", curchar), 15)
                display_text_image({6, 50}, {{curchar, 15}})
            end if
        elsif k = '-' or k = '_' or k = KEY_PGDN or k = KEY_ALT_LF then
            if curchar != 0 then
                CharSet[curchar + 1] = pattern_to_binary(cc)
                curchar = curchar - 1
                cc = EditChars_makepattern(curchar)
                EditChars_drawchar(cc)
                bk_color(1)
                text_at(6, 25, "Current character: " & sprintf("%3d", curchar), 15)
                display_text_image({6, 50}, {{curchar, 15}})
            end if
        elsif k = KEY_ALT_UP then
            if curchar > 31 then
                curchar = curchar - 32
                CharSet[curchar + 1] = pattern_to_binary(cc)
                curchar = curchar - 1
                cc = EditChars_makepattern(curchar)
                EditChars_drawchar(cc)
                bk_color(1)
                text_at(6, 25, "Current character: " & sprintf("%3d", curchar), 15)
                display_text_image({6, 50}, {{curchar, 15}})
            end if
        elsif k = KEY_ALT_DN then
            if curchar < 225 then
                curchar = curchar + 32
                CharSet[curchar + 1] = pattern_to_binary(cc)
                curchar = curchar - 1
                cc = EditChars_makepattern(curchar)
                EditChars_drawchar(cc)
                bk_color(1)
                text_at(6, 25, "Current character: " & sprintf("%3d", curchar), 15)
                display_text_image({6, 50}, {{curchar, 15}})
            end if
        elsif k = 'V' or k = 'v' then
            bufcc = cc
            for lx = 1 to 14 do
                bufcc[lx] = cc[(14 + 1) - lx]
            end for
            cc = bufcc
            CharSet[curchar + 1] = pattern_to_binary(cc)
            EditChars_drawchar(cc)
            mod_flag = 1
        elsif k = 'H' or k = 'h' then
            bufcc = cc
            for lx = 1 to FONT_HEIGHT do
                for ly = 1 to FONT_WIDTH do
                    bufcc[lx][ly] = cc[lx][(FONT_WIDTH + 1) - ly]
                end for
            end for
            cc = bufcc
            CharSet[curchar + 1] = pattern_to_binary(cc)
            EditChars_drawchar(cc)
            mod_flag = 1
        elsif k = KEY_CT_UP then
            bufcc = cc
            for lx = 1 to FONT_HEIGHT - 1 do
                bufcc[lx] = cc[lx + 1]
            end for
            bufcc[FONT_HEIGHT] = cc[1]
            cc = bufcc
            CharSet[curchar + 1] = pattern_to_binary(cc)
            EditChars_drawchar(cc)
            mod_flag = 1
        elsif k = KEY_CT_DN then
            bufcc = cc
            for lx = 2 to FONT_HEIGHT do
                bufcc[lx] = cc[lx - 1]
            end for
            bufcc[1] = cc[FONT_HEIGHT]
            cc = bufcc
            CharSet[curchar + 1] = pattern_to_binary(cc)
            EditChars_drawchar(cc)
            mod_flag = 1
        elsif k = KEY_CT_LF then
            bufcc = cc
            for lx = 1 to FONT_HEIGHT do
                for ly = 1 to FONT_WIDTH - 1 do
                    bufcc[lx][ly] = cc[lx][ly + 1]
                end for
            end for
            for lx = 1 to FONT_HEIGHT do
                bufcc[lx][FONT_WIDTH] = cc[lx][1]
            end for
            cc = bufcc
            CharSet[curchar + 1] = pattern_to_binary(cc)
            EditChars_drawchar(cc)
            mod_flag = 1
        elsif k = KEY_CT_RT then
            bufcc = cc
            for lx = 1 to FONT_HEIGHT do
                for ly = 2 to FONT_WIDTH do
                    bufcc[lx][ly] = cc[lx][ly - 1]
                end for
            end for
            for lx = 1 to FONT_HEIGHT do
                bufcc[lx][1] = cc[lx][FONT_WIDTH]
            end for
            cc = bufcc
            CharSet[curchar + 1] = pattern_to_binary(cc)
            EditChars_drawchar(cc)
            mod_flag = 1
        end if
        
        if x != ox or y != oy or od != draw then
            if x < 1 then
                x = 1
            end if
            if y < 1 then
                y = 1
            end if
            if x > 14 then
                x = 14
            end if
            if y > FONT_WIDTH then
                y = FONT_WIDTH
            end if
        
            t = 50001
            if draw = 1 then
                cc[x][y] = 'x'
                mod_flag = 1
            end if
            if cc[ox][oy] = 'x' then
                display_text_image({4 + ox, 4 + oy*2}, {"ۇۇ"})
            else display_text_image({4 + ox, 4 + oy*2}, {""})
            end if
            
        end if
        
        if mod_flag = 1 then
            font(curchar, cc)
        end if

        t = t + 1
        if t > 10000 then
            t = 0
            if ct = 0 then
                display_text_image({4 + x, 4 + y*2}, {"" & 14 & "" & 14})
                ct = 1
            else
                if cc[x][y] = 'x' then
                    display_text_image({4 + x, 4 + y*2}, {"ۇۇ"})
                else display_text_image({4 + x, 4 + y*2}, {""})
                end if
                ct = 0
            end if
        end if
        
    end while
    return
end procedure

atom inc

CharSet = default_charset
WholeSet = repeat(repeat(0, 64), 8)

inc = 0
for z = 1 to 8 do
    for a = 1 to 32 do
        WholeSet[z][(a * 2) - 1] = inc
        WholeSet[z][a * 2] = 7
        inc = inc + 1
    end for
end for

--EditChars()

